from datetime import datetime
from aitana.tilde import tilde_request


def test_tilde_request():
    """
    Test that the returned dataframe is consistent with 
    the data returned by the tilde exploration tool (https://tilde.geonet.org.nz/ui/data-discovery).
    """
    start_date = '2019-01-01'
    end_date = '2019-01-03'
    df = tilde_request(start_date=datetime(2019, 1, 1), end_date=datetime(2019, 1, 3),
                       domain="envirosensor",
                       name="lake-temperature",
                       station="RU001", sensor="01")
    assert len(df) == 70



import importlib
import os
from typing import Optional
from .tilde import tilde_request


def get_data(filename: Optional[os.PathLike] = None) -> str:
    """Get data from the Aitana dataset.

    Parameters
    ----------
    filename : str, optional
        Name of the file to read. If None, the whole dataset is returned.

    Returns
    -------
    pd.DataFrame
        The requested data.

    Examples
    --------
    >>> from aitana import get_data
    >>> get_data()
    """
    f = importlib.resources.files(__package__)
    return str(f) if filename is None else str(os.path.join(f, filename))



from datetime import timezone, datetime, timedelta
import io
import requests
import pandas as pd


def tilde_request(
    start_date: datetime,
    end_date: datetime,
    domain: str,
    name: str,
    station: str,
    method: str = '-',
    sensor: str = '-',
    aspect: str = '-',
) -> pd.DataFrame:
    """
    Request data from the tilde API (https://tilde.geonet.org.nz/v3/api-docs/).
    See the tilde discovery tool for more information:
    https://tilde.geonet.org.nz/ui/data-discovery/

    Parameters
    ----------
    domain : str
        The domain of the data (e.g. 'manualcollect')
    name : str
        The name of the data (e.g. 'plume-SO2-gasflux')
    station : str
        The station code (e.g. 'WI000')
    method : str
        The method of the data (e.g. 'contouring')
    sensor : str
        The sensor of the data (e.g. 'MC01')
    aspect : str
        The aspect of the data (e.g. 'nil')
    start_date : date
        The start date of the data
    end_date : date
        The end date of the data

    Returns
    -------
    pd.DataFrame
        A pandas dataframe with the requested data
    """
    tilde_url = "https://tilde.geonet.org.nz/v4/data"
    # split the request into historic and latest data
    get_historic = True
    get_latest = False
    start_date = start_date.astimezone(timezone.utc)
    end_date = end_date.astimezone(timezone.utc)
    _tstart = str(start_date.date())
    _tend = str(end_date.date())
    _today = datetime.now(timezone.utc).date()
    if end_date.date() > (_today - timedelta(days=29)):
        _tend = str((end_date.date() - timedelta(days=29)))
        get_latest = True
    if start_date.date() > (_today - timedelta(days=29)):
        get_historic = False

    assert get_historic or get_latest, "Check start and end dates."

    if get_latest:
        latest = f"{tilde_url}/{domain}/{station}/{name}/{sensor}/{method}/{aspect}/latest/30d"
        rval = requests.get(latest, headers={"Accept": "text/csv"})
        if rval.status_code != 200:
            msg = f"Download of {name} for {station} failed with status code {rval.status_code}"
            msg += f" and url {latest}"
            raise ValueError(msg)
        df_latest = pd.read_csv(
            io.StringIO(rval.text),
            index_col="timestamp",
            parse_dates=["timestamp"],
            usecols=["timestamp", "value", "error"],
            date_format="ISO8601",
        )
    if get_historic:
        historic = f"{tilde_url}/{domain}/{station}/{name}/{sensor}/{method}/{aspect}/"
        historic += f"{_tstart}/{_tend}"
        rval = requests.get(historic, headers={"Accept": "text/csv"})
        if rval.status_code != 200:
            msg = f"Download of {name} for {station} failed with status code {rval.status_code}"
            msg += f" and url {historic}"
            raise ValueError(msg)
        data = io.StringIO(rval.text)
        df_historic = pd.read_csv(
            data,
            index_col="timestamp",
            parse_dates=["timestamp"],
            usecols=["timestamp", "value", "error"],
            date_format="ISO8601",
        )
        if get_latest and len(df_latest) > 0:
            df = df_historic.combine_first(df_latest)
        else:
            df = df_historic
    else:
        df = df_latest
    df.rename(columns={"value": "obs", "error": "err"}, inplace=True)
    df.index.name = "dt"
    return df.loc[start_date:end_date.replace(hour=23, minute=59, second=59)]



from datetime import datetime, timezone
import os

import numpy as np
import pandas as pd
import requests

from aitana import get_data
from aitana.tilde import tilde_request
from aitana.assimilate import SO2FusionModel
from aitana.util import cache_dataframe, eqRate, reindex
from aitana.wfs import wfs_request


class Gas(object):

    def __init__(self, start_date: datetime, end_date: datetime):
        self.start_date = pd.to_datetime(start_date, utc=True)
        self.end_date = pd.to_datetime(end_date, utc=True)

    @cache_dataframe()
    def so2(self, fuse: bool = False, smooth: bool = True) -> pd.DataFrame:

        dataframes = {}
        method = "cospec"
        if self.end_date > datetime(2023, 11, 9, 23, 1, 0, tzinfo=timezone.utc):
            method = "contouring"
        try:
            df = tilde_request(start_date=datetime(2000, 1, 1),
                               end_date=self.end_date,
                               domain="manualcollect",
                               name="plume-SO2-gasflux",
                               method=method,
                               station="WI000",
                               sensor="MC01")
        except ValueError:
            df = pd.DataFrame()
        else:
            dataframes['WI000'] = df

        if not fuse:
            return df * 86.4

        # Note: north-east point (WID01) is reporting consistently too low values
        for station in ["WID01", "WID02"]:
            try:
                df_ = tilde_request(start_date=datetime(2000, 1, 1),
                                    end_date=self.end_date,
                                    domain="scandoas",
                                    name="gasflux",
                                    method="reviewed",
                                    station=station,
                                    sensor="01")
            except ValueError:
                continue
            else:
                dataframes[station] = df_

        startdate = np.array([df.index.min()
                             for df in list(dataframes.values())]).min()
        enddate = np.array([df.index.max()
                           for df in list(dataframes.values())]).max()
        dates = pd.date_range(startdate, enddate, freq="1D")
        gasdata = {}
        interval = "1D"
        for key, df in dataframes.items():
            gasdata[key] = (
                df["obs"].groupby(pd.Grouper(freq=interval)
                                  ).median().reindex(dates).values
            )

        gasdata = pd.DataFrame(gasdata, index=dates)

        # define a covariance matrix for the observations that gives cospec values
        # a higher weight than mdoas values
        obs_cov = np.diag(np.array([5., 200., 20.])**2)
        model = SO2FusionModel(measurements=gasdata, initial_state=np.array([0]),
                               initial_cov=np.diag([[1]]),
                               obs_cov=obs_cov, k_states=1, k_posdef=1)
        # Set initial parameters for process noise
        initial_params = [0.5]
        # Fit the model
        result = model.fit(start_params=initial_params, disp=False)

        # Get the smoothed state estimates (filtered values)
        filtered_state = result.filtered_state[0]
        smoothed_state = result.smoothed_state[0]

        mean_trace = filtered_state
        error = np.sqrt(result.filtered_state_cov[0, 0])
        if smooth:
            mean_trace = smoothed_state
            error = np.sqrt(result.smoothed_state_cov[0, 0])
        so2df = pd.DataFrame(
            {"obs": mean_trace, "err": error}, index=gasdata.index)
        so2df.index.name = "dt"
        return so2df * 86.4

    @cache_dataframe()
    def co2(self) -> pd.DataFrame:
        df = tilde_request(start_date=datetime(2000, 1, 1),
                           end_date=self.end_date,
                           domain="manualcollect",
                           name="plume-CO2-gasflux",
                           method="contouring",
                           station="WI000", sensor="MC01")
        return df * 86.4

    @cache_dataframe()
    def h2s(self) -> pd.DataFrame:
        df = tilde_request(start_date=datetime(2000, 1, 1),
                           end_date=self.end_date,
                           domain="manualcollect",
                           name="plume-H2S-gasflux",
                           method="contouring",
                           station="WI000", sensor="MC01")
        return df * 86.4


class Seismicity(object):

    def __init__(self, start_date: datetime, end_date: datetime):
        self.start_date = pd.to_datetime(start_date, utc=True)
        self.end_date = pd.to_datetime(end_date, utc=True)

    def rsam(self) -> pd.DataFrame:
        """
        Load RSAM values NZ.WIZ.10.HHZ.The first time you call
        this it will download the data from Zenodo and store it
        locally.

        Returns:
        --------
            :param df: Dataframe with RSAM values.
            :type df: :class:`pandas.DataFrame`
        """
        rsam_fn = get_data('data/RSAM_NZ.WIZ.10.HHZ.csv')
        if not os.path.isfile(rsam_fn):
            # download the data from zenodo
            print('Downloading RSAM data from Zenodo')
            url = "https://zenodo.org/records/14759090/files/RSAM_NZ.WIZ.10.HHZ.csv"
            response = requests.get(url)
            if response.status_code == 200:
                with open(rsam_fn, "wb") as f:
                    f.write(response.content)
                print('Successfully downloaded RSAM data from Zenodo')
            else:
                print('Failed to download RSAM data from Zenodo')

        df = pd.read_csv(rsam_fn, skiprows=1, parse_dates=True, names=['dt', 'obs'], index_col=0,
                         date_format='ISO8601')
        if self.end_date.tzinfo is not None:
            df.index = df.index.tz_localize(timezone.utc)
        df = df[df.index <= str(self.end_date)]
        df = df[df.index >= str(self.start_date)]
        return df

    @cache_dataframe()
    def quakes(self):
        center_point = "177.186833+-37.523118"
        return wfs_request(self.start_date, self.end_date, radius=20000,
                           center_point=center_point, maxdepth=30)


@cache_dataframe()
def eruptions(min_size: int = 0, dec_interval: str = None) -> pd.DataFrame:
    """
    This function loads the eruption catalogue for White Island and
    declusters it.

    :param eruption_scale: The minimum eruption scale (between 0 and 4)
                           to use. This is currently ignored as the
                           White Island catalogue does not contain
                           eruption scale.
    :type eruption_scale: int
    :param dec_interval: The declustering interval, which is the minimum
                         distance in time between any two eruptions.
    :type dec_interval: :class:`pandas.DateOffset` or str
    :param datadir: Path that contains the catalogue file in csv format.
    :type datadir: str
    :returns: Declustered eruption catalogue
    :rtype: :class:`pandas.DataFrame`
    """
    eruptions = pd.read_csv(
        get_data("data/WhiteIs_Eruption_Catalogue.csv"),
        index_col="Date",
        parse_dates=True,
        comment="#",
    )

    # Select eruptions of a particular scale or larger
    eruptions = eruptions[eruptions.Activity_Scale >= min_size].copy()

    if dec_interval is not None:
        # duplicate time index as a data column
        eruptions.insert(1, "tvalue", eruptions.index)
        # calculate intereruption event time
        delta = (eruptions["tvalue"] - eruptions["tvalue"].shift()).fillna(
            pd.Timedelta(seconds=0)
        )
        eruptions.insert(1, "delta", delta)
        eruptions.iloc[0, 1] = pd.Timedelta(dec_interval)
        eruptions = eruptions[(eruptions.delta >= dec_interval)]
        eruptions = eruptions.drop(columns=["tvalue", "delta"])

    eruptions.index = pd.DatetimeIndex(eruptions.index)
    eruptions = eruptions.tz_localize('utc')
    return eruptions


@cache_dataframe()
def load_all(
    fill_method="interpolate",
    start_date=datetime(2005, 1, 1, tzinfo=timezone.utc),
    end_date=datetime.now(timezone.utc),
    ignore_data=(),
    fuse_so2=False,
):
    cols = {}
    s = Seismicity(start_date, end_date)
    cat = s.quakes()
    dft = eqRate(cat, fixed_time=7).resample("D").mean().interpolate()
    new_dates = pd.date_range(dft.index[0], end_date, freq="D")
    if "Eqr" not in ignore_data:
        cols["Eqr"] = reindex(dft, new_dates, fill_method=fill_method)
    if "RSAM" not in ignore_data:
        rsam = s.rsam()
        cols["RSAM"] = reindex(rsam, new_dates, fill_method=fill_method)

    g = Gas(start_date, end_date)
    if "CO2" not in ignore_data:
        co2 = g.co2()
        cols["CO2"] = reindex(co2, new_dates, fill_method=fill_method)
    if "H2S" not in ignore_data:
        h2s = g.h2s()
        cols["H2S"] = reindex(h2s, new_dates, fill_method=fill_method)
    if "SO2" not in ignore_data:
        smooth = False
        if fill_method == "interpolate":
            smooth = True
            fuse_so2 = True
        so2 = g.so2(fuse=fuse_so2, smooth=smooth)
        cols["SO2"] = reindex(so2, new_dates, fill_method=fill_method)
    rdf = pd.DataFrame(cols, index=new_dates)
    return rdf



import numpy as np
import pandas as pd
from aitana.whakaari import Gas, Seismicity, eruptions, load_all


def test_gas():
    g = Gas(start_date='2020-01-01', end_date='2024-12-31')
    df1 = g.so2(fuse=False, smooth=False)
    df2 = g.so2(fuse=True, smooth=True)
    assert (df1.obs.max() / 86.4 - 31.9) < 0.001
    assert (df1.obs.min() / 86.4 - 0.355) < 0.001
    # test that df2 is smoother than df1 by comparing standard deviations
    assert df2.obs.std() < df1.obs.std()
    df = g.co2()
    assert df.obs.loc['2023-11-09'].values[0] / 86.4 - 5.2 < 0.001

    df = g.h2s()
    assert df.obs.loc['2024-02-29'].values[0] / 86.4 - 0.29 < 0.001


def test_seismicity():
    s = Seismicity('2019-12-01', '2020-01-31')
    df = s.rsam()
    assert df.obs.max() > 7900
    dfq = s.quakes()
    assert dfq.index[0] >= pd.to_datetime('2019-12-01', utc=True)
    assert dfq.index[-1] <= pd.to_datetime('2020-01-31T23:59:59', utc=True)
    assert dfq.magnitude.max() > 4.
    assert dfq.magnitude.min() < 2.


def test_eruptions():
    df = eruptions(end_date='2024-12-31')
    assert np.diff(df.index).min().days == 0
    df = eruptions(dec_interval='10D', end_date='2024-12-31')
    assert np.diff(df.index).min().days == 10
    df = eruptions(start_date='2010-01-01', end_date='2024-12-31', min_size=2)
    assert df.shape[0] == 4


def test_load_all():
    end_date = '2025-02-14'
    df = load_all(end_date=end_date)
    assert df.iloc[-1].isnull().sum() == 0



from datetime import datetime
import pandas as pd


def wfs_request(
    start_date: datetime,
    end_date: datetime,
    polygon: str = None,
    radius: int = None,
    center_point: str = None,
    maxdepth: int = 30
) -> pd.DataFrame:
    """
    Load volcano-tectonic earthquake catalogues from GeoNet's WFS service (http://wfs.geonet.org.nz).

    Parameters:
    -----------
        polygon : str
            A string with the polygon coordinates in the format
            "lon1+lat1,lon2+lat2,...,lonN+latN,lon1+lat1"
        radius : int
            The radius around the volcano summit in meters.
        center_point : str
            The center point of the radius in the format "lon+lat".
        maxdepth : int
            The maximum depth of the earthquakes in kilometers.

    Returns:
    --------
        pandas.DataFrame
            A dataframe with the earthquake catalogue.
    """
    wfs_url = "http://wfs.geonet.org.nz/geonet/ows?service=WFS&version=1.0.0"
    if polygon is None and radius is None:
        raise ValueError("Either polygon or radius must be provided.")
    start_date = start_date.strftime("%Y-%m-%dT%H:%M:%S.0Z")
    url = wfs_url
    url += (
        "&request=GetFeature&typeName=geonet:quake_search_v1&outputFormat=csv"
    )
    url += f"&cql_filter=origintime>={start_date}"

    if radius is not None:
        url += (
            f"+AND+DWITHIN(origin_geom,Point+({center_point}),{radius},meters)+AND+depth<{maxdepth}"
        )
    if polygon is not None:
        url += f"+AND+WITHIN(origin_geom,POLYGON(({polygon})))+AND+depth<{maxdepth}"

    cat = pd.read_csv(url, parse_dates=["origintime"], index_col="origintime")
    cat.sort_index(inplace=True)
    if cat.size == 0:
        msg = "There are no earthquakes available for"
        msg += f"the selected date range ({str(start_date)}, {str(end_date)}) in "
        if radius is not None:
            msg += "a radius of {} m".format(radius)
        if polygon is not None:
            msg += "the selected region: {}".format(polygon)
        raise ValueError(msg)

    return cat



import pandas as pd
import numpy as np

from aitana.util import cache_dataframe


class DataFetcher:
    def __init__(self, start_date, end_date):
        self.start_date = start_date
        self.end_date = end_date

    @cache_dataframe()
    def fetch_data(self):
        # Simulate fetching data for the specified date range
        date_range = pd.date_range(
            start=pd.to_datetime('2024-01-01', utc=True), end=self.end_date, freq="D")
        data = pd.DataFrame({
            "value": np.random.rand(len(date_range))
        }, index=date_range)
        return data


@cache_dataframe()
def fetch_data_function(val, end_date):
    date_range = pd.date_range(
        start=pd.to_datetime("2024-01-01", utc=True), end=end_date, freq="D")
    data = pd.DataFrame(
        {"value": np.random.rand(len(date_range)) * val}, index=date_range)
    return data


def test_cache():
    # Usage
    fetcher = DataFetcher(start_date="2024-01-01", end_date="2024-01-10")

    # Fetch using instance attributes
    df1 = fetcher.fetch_data(clear_cache=True)
    assert len(df1) == 10

    # Update instance attributes
    fetcher.start_date = "2024-01-10"
    fetcher.end_date = "2024-01-15"

    # Fetch using updated attributes
    df2 = fetcher.fetch_data()
    assert len(df2) == 6

    # Cached data is updated
    assert len(fetcher.fetch_data.cached_df) == 15

    # Fetch using function arguments
    df3 = fetch_data_function(5, start_date="2024-02-01",
                              end_date="2024-02-10", clear_cache=True)
    assert len(df3) == 10

    # Update function arguments
    df4 = fetch_data_function(
        4, start_date="2024-02-10", end_date="2024-02-15")
    assert len(df4) == 6

    # Cached data is updated
    assert len(fetch_data_function.cached_df) == 46



from datetime import datetime, timezone, timedelta

import pytest

from aitana.ruapehu import CraterLake, Gas, Seismicity, eruptions, load_all


def test_crater_lake():
    tstart = datetime(2023, 1, 1)
    tend = datetime(2023, 1, 2)
    rcl = CraterLake(tstart, tend)
    df = rcl.temperature()
    assert df.obs.loc['2023-01-01'] - 30.908 < 0.001

    tstart = datetime(2022, 12, 21)
    tend = datetime(2022, 12, 22)
    rcl = CraterLake(tstart, tend)
    df = rcl.water_analyte('Mg')
    assert df.obs.loc['2022-12-21'] - 362.0 < 0.001

    tstart = datetime(2010, 11, 18)
    tend = datetime(2010, 11, 19)
    rcl = CraterLake(tstart, tend)
    df = rcl.water_level()
    assert df.obs.loc['2010-11-18'].mean() - 1.241 < 0.001

    tstart = datetime(2021, 5, 8)
    tend = datetime(2021, 5, 9)
    rcl = CraterLake(tstart, tend)
    df = rcl.water_level()
    assert df.obs.loc['2021-05-08'].mean() - 1.269 < 0.001


def test_gas():
    tstart = datetime(2023, 12, 20)
    tend = datetime(2023, 12, 21)
    g = Gas(tstart, tend)
    df = g.so2()
    assert df.obs.loc['2023-12-20T21:30:00'] - 132.192 < 0.001

    df = g.co2()
    assert df.obs.loc['2023-12-20T21:30:00'] - 851.904 < 0.001

    df = g.h2s()
    assert df.obs.loc['2023-12-20T21:30:00'] == 0.

    tstart = datetime(2021, 1, 1)
    tend = datetime(2021, 12, 31)
    g = Gas(tstart, tend)
    df = g.so2()
    assert df.shape == (5, 2)

    with pytest.raises(ValueError):
        Gas(datetime(2021, 10, 1), datetime(2021, 10, 3)).so2()


def test_seismicity():
    tstart = datetime(2023, 12, 1)
    tend = datetime(2023, 12, 31)
    g = Seismicity(tstart, tend)
    cat1 = g.regional()
    cat2 = g.cone()
    cat3 = g.rm_duplicates(cat1, cat2)
    assert cat1.shape == (43, 20)
    assert cat2.shape == (10, 20)
    assert cat3.shape == (33, 20)


def test_eruptions():
    e1 = eruptions(min_size=3,
                   end_date=datetime(2024, 11, 30, tzinfo=timezone.utc))
    e2 = eruptions(min_size=3, dec_interval='14D',
                   end_date=datetime(2024, 11, 30, tzinfo=timezone.utc))
    assert e1.shape == (293, 5)
    assert e2.shape == (57, 5)


def test_rsam():
    tstart = datetime(2007, 1, 1)
    tend = datetime(2024, 11, 30)
    g = Seismicity(tstart, tend)
    rsam = g.daily_rsam()
    assert rsam.shape == (6159, 1)

    tstart = datetime(2023, 11, 1, tzinfo=timezone.utc)
    tend = datetime(2023, 11, 30, tzinfo=timezone.utc)
    g = Seismicity(tstart, tend)
    rsam = g.daily_rsam()
    assert rsam.shape == (30, 1)

    # test local rsam server
    # tstart = datetime(2023, 11, 1, tzinfo=timezone.utc)
    # tend = datetime.now(timezone.utc)
    # tstart = tend - timedelta(days=30)
    # g = Seismicity(tstart, tend)
    # rsam = g.daily_rsam(update=True)
    # print(rsam)


def test_load_all():
    end_date = datetime(2024, 11, 30)
    df = load_all(end_date=end_date)
    assert df.iloc[-1].isnull().sum() == 0



import os
from datetime import datetime, timezone
import pandas as pd
import numpy as np

from aitana.tilde import tilde_request
from aitana.wfs import wfs_request
from aitana import get_data
from aitana.util import cache_dataframe, gradient, reindex, eqRate


class CraterLake(object):

    def __init__(self, start_date: datetime, end_date: datetime):
        self.start_date = start_date
        self.end_date = end_date

    @cache_dataframe()
    def temperature(self, resample: str = "D", interpolate: str = None,
                    exclude1995: bool = True, dropna: bool = True) -> pd.DataFrame:
        """
        Read crater lake temperature from Tilde (https://tilde.geonet.org.nz).

        Parameters:
        -----------
            :param resample: Average over resampling interval and
                            linearly interpolate in between. The
                            interval should be one of 'D', 'W',
                            or 'M'
            :type resample: str
            :param interpolate: Interpolate between resampled points.
                                Only takes effect if resample is not
                                None.
            :type interpolate: str

        Returns:
        --------
            :param df: A dataframe with the temperature time series.
            :type df: :class:`pandas.Dataframe`
        """
        dataframes = []
        # read in datalogger temperature data
        try:
            df = tilde_request(start_date=datetime(2019, 5, 21, 4, 30, 0), end_date=self.end_date,
                               domain="envirosensor",
                               name="lake-temperature",
                               station="RU001", sensor="04")
        except ValueError:
            pass
        else:
            dataframes.append(df)

        try:
            df1 = tilde_request(start_date=datetime(1993, 12, 11), end_date=self.end_date,
                                domain="envirosensor",
                                name="lake-temperature",
                                station="RU001", sensor="01")
        except ValueError:
            pass
        else:
            dataframes.append(df1)

        # Read in the manual temperatures
        # water analysis lab, thermometer
        try:
            dfmc1 = tilde_request(start_date=datetime(1954, 2, 14), end_date=self.end_date,
                                  domain="manualcollect",
                                  name="lake-temperature",
                                  station="RU001",
                                  method="thermometer",
                                  sensor="MC01")
        except ValueError:
            pass
        else:
            if len(dfmc1) > 0:
                dataframes.append(dfmc1)

        # water analysis lab, thermocouple
        try:
            dfmc1_2 = tilde_request(start_date=datetime(1991, 1, 13), end_date=self.end_date,
                                    domain="manualcollect",
                                    name="lake-temperature",
                                    station="RU001",
                                    method="thermocouple",
                                    sensor="MC01")
        except ValueError:
            pass
        else:
            if len(dfmc1_2) > 0:
                dataframes.append(dfmc1_2)

        # thermocouple 1
        try:
            dfmc3 = tilde_request(start_date=datetime(1998, 3, 17), end_date=self.end_date,
                                  domain="manualcollect",
                                  name="lake-temperature",
                                  station="RU001",
                                  method="thermocouple",
                                  sensor="MC03")
        except ValueError:
            pass
        else:
            if len(dfmc3) > 0:
                dataframes.append(dfmc3)

        # thermocouple 2
        try:
            dfmc4 = tilde_request(start_date=datetime(1991, 1, 13), end_date=self.end_date,
                                  domain="manualcollect",
                                  name="lake-temperature",
                                  station="RU001",
                                  method="thermocouple",
                                  sensor="MC04")
        except ValueError:
            pass
        else:
            if len(dfmc4) > 0:
                dataframes.append(dfmc4)

        if len(dataframes) == 0:
            raise ValueError(
                f"No data found for lake temperature between {self.start_date} and {self.end_date}")

        df = dataframes[0]
        if len(dataframes) > 1:
            for df1 in dataframes[1:]:
                df = df.combine_first(df1)

        if resample is not None:
            df = df.resample("D").mean()
            if interpolate is not None:
                df = df.interpolate(interpolate)

        # remove observations following the 1995 eruption when
        # the measured temperatures do not represent a lake.
        if exclude1995:
            cond1 = df.index <= "1995-09-20 00:00:00"
            cond2 = df.index >= "2000-01-01 00:00:00"
            df = df[(cond1) | (cond2)]
        # get rid of nans
        if dropna:
            df.dropna(inplace=True)
        # change annoying temp label and drop old column
        df.rename(columns={" t (C)": "obs"}, inplace=True)
        return df

    @cache_dataframe()
    def water_level(self):
        """
        Read crater lake water level from Tilde (https://tilde.geonet.org.nz).
        """
        sensors = ['02', '03']
        dataframes = []
        for sensor in sensors:
            try:
                df = tilde_request(start_date=datetime(2009, 4, 15, 2, 0, 0), end_date=self.end_date,
                                   domain="envirosensor",
                                   name="lake-height",
                                   station="RU001", sensor=sensor)
            except ValueError:
                continue
            if len(df) > 0:
                dataframes.append(df)

        df = dataframes[0]
        if len(dataframes) > 1:
            for df1 in dataframes[1:]:
                df = df.combine_first(df1)
        return df

    @cache_dataframe()
    def water_analyte(self, analyte: str, resample: str = None, interpolate: str = None,
                      exclude1995: bool = True) -> pd.DataFrame:
        """
        Download water analyte data from Tilde (https://tilde.geonet.org.nz).

        Parameters:
        -----------
            analyte : str
                The analyte to download. This can be one of
                    'Al, 'As', 'B', 'Br', 'Ca', 'Cl', 'Cs', 'F',
                    'Fe', 'H2S', 'K', 'Li', 'Mg', 'NH3', 'NO3-N', 'Na',
                    'PO4-P', 'Rb', 'SO4', 'SiO2', 'd18O', 'd2H', 'ph'
            resample : str
                Resample the data to a given interval. This can be anything
                allowed by :pandas.DataFrame.resample:. 
            interpolate : str
                Interpolate between points. This can be anything allowed by
                :pandas.DataFrame.interpolate:.
            exclude1995 : bool
                Exclude data from 1995 when there wasn't a lake.
        Returns:
        --------
            pandas.DataFrame
                A dataframe with the water analyte time series.
        """
        dataframes = []
        sensors = ["MC03", "MC04", "MC01"]
        for sensor in sensors:
            try:
                df = tilde_request(start_date=datetime(1964, 5, 9),
                                   end_date=self.end_date,
                                   domain="manualcollect",
                                   name=f"lake-{analyte}-conc",
                                   station="RU001",
                                   sensor=sensor)
            except ValueError:
                continue
            if len(df) > 0:
                dataframes.append(df)
        if len(dataframes) == 0:
            raise ValueError(
                f"No data found for {analyte} between {self.start_date} and {self.end_date}")
        df = dataframes[0]
        if len(dataframes) > 1:
            for df1 in dataframes[1:]:
                df = df.combine_first(df1)
        if resample is not None:
            df = df.resample(resample).mean()

        if interpolate is not None:
            df = df.interpolate(interpolate)

        if exclude1995:
            cond1 = df.index <= "1995-09-20 00:00:00"
            cond2 = df.index >= "2000-01-01 00:00:00"
            df = df[(cond1) | (cond2)]
        return df


class Gas(object):

    def __init__(self, start_date: datetime, end_date: datetime):
        self.start_date = start_date
        self.end_date = end_date

    @cache_dataframe()
    def so2(self) -> pd.DataFrame:
        dataframes = []
        for method in ["cospec", "contouring", "flyspec", "mobile-doas"]:
            try:
                df = tilde_request(start_date=datetime(2003, 5, 27),
                                   end_date=self.end_date,
                                   domain="manualcollect",
                                   name="plume-SO2-gasflux",
                                   method=method,
                                   station="RU000", sensor="MC01")
            except ValueError:
                continue
            if len(df) > 0:
                dataframes.append(df)
        if len(dataframes) == 0:
            raise ValueError(
                f"No data found for SO2 between {self.start_date} and {self.end_date}")
        df = dataframes[0]
        if len(dataframes) > 1:
            for df1 in dataframes[1:]:
                df = df.combine_first(df1)
        df['obs'] *= 86.4
        df['err'] *= 86.4
        return df

    @cache_dataframe()
    def co2(self) -> pd.DataFrame:
        df = tilde_request(start_date=max(self.start_date, datetime(2003, 5, 27)),
                           end_date=self.end_date,
                           domain="manualcollect",
                           name="plume-CO2-gasflux",
                           method="contouring",
                           station="RU000", sensor="MC01")
        df['obs'] *= 86.4
        df['err'] *= 86.4
        return df

    @cache_dataframe()
    def h2s(self) -> pd.DataFrame:
        df = tilde_request(start_date=max(self.start_date, datetime(2004, 4, 21)),
                           end_date=self.end_date,
                           domain="manualcollect",
                           name="plume-H2S-gasflux",
                           method="contouring",
                           station="RU000", sensor="MC01")
        df['obs'] *= 86.4
        df['err'] *= 86.4
        return df


class Seismicity(object):

    def __init__(self, start_date: datetime, end_date: datetime):
        self.start_date = start_date
        self.end_date = end_date

    @cache_dataframe()
    def regional(self):
        """
        Load an earthquake catalogue for an area containing Waiouru and National Park.
        """

        polygon = (
            "175.32+-39.50,+175.32+-39.18,+175.77+-39.18,+175.77+-39.50,+175.32+-39.50"
        )
        return wfs_request(self.start_date, self.end_date, polygon=polygon)

    @cache_dataframe()
    def cone(self):
        """
        Load an earthquake catalogue for an area of 7 km around the summit.
        """
        radius = 7000
        center_point = "175.57+-39.28"
        return wfs_request(self.start_date, self.end_date, radius=radius, center_point=center_point)

    def rm_duplicates(self, cat1: pd.DataFrame, cat2: pd.DataFrame) -> pd.DataFrame:
        """
        Remove events from catalogue cat1 that are also in cat2 and return a new catalogue.

        Parameters:
        -----------
            cat1 : pandas.DataFrame
                The first catalogue.
            cat2 : pandas.DataFrame
                The second catalogue.

        Returns:
        --------
            pandas.DataFrame
                A new catalogue with the events from cat1 that are not in cat2.
        """
        idxs = []
        cat_tmp = cat1.copy()
        for i in range(cat_tmp.shape[0]):
            if cat_tmp.iloc[i].publicid in cat2.publicid.values:
                idxs.append(i)
        dates_to_drop = cat_tmp.index[idxs]
        cat_tmp.drop(dates_to_drop, inplace=True)
        return cat_tmp

    def daily_rsam(self, update=False) -> pd.DataFrame:
        """
        Load RSAM values from DRZ, MAVZ and FWVZ combined by scaling
        MAVZ and FWVZ RSAM values with DRZ.

        Returns:
        --------
            :param df: Dataframe with RSAM values.
            :type df: :class:`pandas.DataFrame`
        """
        df = pd.read_csv(get_data("data/ruapehu_rsam.csv"),
                         parse_dates=True, index_col=0)
        if self.end_date.tzinfo is not None:
            df.index = df.index.tz_localize(timezone.utc)
        if self.end_date > df.index[-1] and update:
            df_drz = pd.read_csv(
                get_data("data/DRZ_scaled_MAVZ.csv"), parse_dates=True, index_col=0
            )
            df_drz.rename(columns={"RSAM": "obs"}, inplace=True)
            url = "http://kaizen.gns.cri.nz:9157/feature?name=rsam"
            url += "&starttime=2007-01-01T00:00:00"
            url += f"&endtime={self.end_date.isoformat().split('+')[0]}"
            url += "&volcano=Ruapehu&site=FWVZ"
            try:
                df_fwvz = pd.read_csv(
                    url,
                    parse_dates=True,
                    index_col=0,
                    date_format="ISO8601",
                )
            except Exception as e:
                print(url)
                raise e
            df_fwvz["rsam"] = np.where(
                df_fwvz["rsam"] > 1e30, np.nan, df_fwvz["rsam"])
            df_fwvz_daily = df_fwvz.resample("1D").mean()
            df_fwvz_daily_scaled = -2.3945 + 3.5062 * df_fwvz_daily
            df_fwvz_daily_scaled.rename(columns={"rsam": "obs"}, inplace=True)
            df = df_drz.combine_first(df_fwvz_daily_scaled)
            # remove duplicated dates:
            df = df.loc[~df.index.duplicated(), :]
            df.to_csv(get_data("data/ruapehu_rsam.csv"))
            if self.end_date.tzinfo is not None:
                df.index = df.index.tz_localize(timezone.utc)
        df = df[df.index <= str(self.end_date)]
        df = df[df.index >= str(self.start_date)]
        return df


@cache_dataframe()
def eruptions(min_size: int = 0, dec_interval: str = None) -> pd.DataFrame:
    """
    This function loads the eruption catalogue for Mt. Ruapehu and declusters
    it if required. The catalogue is based on Brad Scott's GNS Science Report.
    Declustering is turned on if dec_interval is greater than zero. This will
    also exclude magmatic eruption episodes between 9 January 1944 and
    8 January 1946 as well as 6 January 1995 to 12 January 1997 as these aren't
    independent events.

    Parameters:
    -----------
        min_size : int
            The minimum eruption size to include in the catalogue.
        dec_interval : str 
            The declustering interval in pandas complient time delta format.

    Returns:
    --------
        pandas.DataFrame
            A dataframe with the (declustered) eruption catalogue.
    """
    filename = "https://raw.githubusercontent.com/GeoNet/data/refs/heads/main/"
    filename += "historic-volcanic-activity/historic_eruptive_activity_ruapehu.csv"
    df = pd.read_csv(filename, index_col=0, parse_dates=True, comment='#')

    # Select eruptions of a particular scale or larger
    scaled_eruptions = df[df['Activity Scale'] >= min_size].copy()

    if dec_interval is not None:
        # duplicate time index as a data column
        scaled_eruptions.insert(1, "tvalue", scaled_eruptions.index)
        # calculate intereruption event time
        delta = (scaled_eruptions["tvalue"] - scaled_eruptions["tvalue"].shift()).fillna(
            pd.Timedelta(seconds=0)
        )
        scaled_eruptions.insert(1, "delta", delta)
        scaled_eruptions.iloc[0, 1] = pd.Timedelta(dec_interval)
        # Exclude certain date ranges from calculations to ensure a more
        # Poisson-like process by removing long term eruption periods.
        period1 = scaled_eruptions.index < "1944-10-01 00:00:00"
        period2 = scaled_eruptions.index > "1946-08-01 00:00:00"
        scaled_eruptions = scaled_eruptions[period1 | period2]
        period3 = scaled_eruptions.index < "1995-07-01 00:00:00"
        period4 = scaled_eruptions.index > "1997-12-01 00:00:00"
        scaled_eruptions = scaled_eruptions[period3 | period4]
        scaled_eruptions = scaled_eruptions[(
            scaled_eruptions.delta >= dec_interval)]
        scaled_eruptions = scaled_eruptions.drop(columns=["delta", "tvalue"])
    scaled_eruptions = scaled_eruptions.tz_localize('utc')
    return scaled_eruptions


@cache_dataframe()
def load_all(
    fill_method="interpolate",
    end_date=datetime.now(timezone.utc),
):
    end_date = end_date.replace(tzinfo=timezone.utc)
    cols = {}

    # load crater lake data
    rcl = CraterLake(start_date=datetime(1954, 1, 1), end_date=end_date)
    dft = rcl.temperature()
    grad = gradient(dft)
    new_dates = pd.date_range(dft.index[0], end_date, freq="D")
    cols["TemperatureBin"] = reindex(dft, new_dates, fill_method=fill_method)
    cols["GradientBin"] = reindex(grad, new_dates, fill_method=fill_method)
    mg = rcl.water_analyte("Mg")
    dmg = gradient(mg, period="90D")
    cols["Mg"] = reindex(dmg, new_dates, fill_method=fill_method)
    so4 = rcl.water_analyte("SO4")
    dso4 = gradient(so4, period="90D")
    cols["SO4"] = reindex(dso4, new_dates, fill_method=fill_method)
    cl = rcl.water_analyte("Cl")
    mg_cl = mg / cl
    dmg_cl = gradient(mg_cl, period="90D")
    cols["Mg_ClBin"] = reindex(dmg_cl, new_dates, fill_method=fill_method)
    na = rcl.water_analyte("Na")
    mg_na = mg / na
    dmg_na = gradient(mg_na, period="90D")
    cols["Mg_Na"] = reindex(dmg_na, new_dates, fill_method=fill_method)
    k = rcl.water_analyte("K")
    mg_k = mg / k
    dmg_k = gradient(mg_k, period="90D")
    cols["Mg_K"] = reindex(dmg_k, new_dates, fill_method=fill_method)
    al = rcl.water_analyte("Al")
    mg_al = mg / al
    dmg_al = gradient(mg_al, period="90D")
    cols["Mg_Al"] = reindex(dmg_al, new_dates, fill_method=fill_method)

    # Load gas data
    rg = Gas(start_date=datetime(1954, 1, 1), end_date=end_date)
    co2 = rg.co2()
    co2 = co2.groupby(pd.Grouper(freq="1D")).median()
    cols["CO2"] = reindex(co2, new_dates, fill_method=fill_method)
    so2 = rg.so2()
    so2 = so2.groupby(pd.Grouper(freq="1D")).median()
    cols["SO2"] = reindex(so2, new_dates, fill_method=fill_method)
    h2s = rg.h2s()
    h2s = h2s.groupby(pd.Grouper(freq="1D")).median()
    cols["H2S"] = reindex(h2s, new_dates, fill_method=fill_method)
    cs = co2 / so2
    dcs = gradient(cs, period="90D")
    cols["CO2_SO2"] = reindex(dcs, new_dates, fill_method=fill_method)

    # Load seismic data
    rs = Seismicity(start_date=datetime(2015, 1, 1), end_date=end_date)
    cat_outer = rs.regional()
    cat_inner = rs.cone()
    cat_outer = rs.rm_duplicates(cat_outer, cat_inner)
    eqr_outer = eqRate(cat_outer, fixed_time=7).resample(
        "D").mean().interpolate()
    eqr_inner = eqRate(cat_inner, fixed_time=7).resample(
        "D").mean().interpolate()
    cols["Eqr_outer"] = reindex(eqr_outer, new_dates, fill_method=fill_method)
    cols["Eqr_inner"] = reindex(eqr_inner, new_dates, fill_method=fill_method)
    rsam = rs.daily_rsam()
    cols["RSAM"] = reindex(rsam, new_dates, fill_method=fill_method)
    rsam100 = rsam.rolling(window="100D", min_periods=1).max()
    cols["RSAM100"] = reindex(rsam100, new_dates, fill_method=fill_method)
    return pd.DataFrame(cols, index=new_dates)



import numpy as np
from statsmodels.tsa.statespace.mlemodel import MLEModel


class SemiLinearTrend(MLEModel):
    def __init__(self, endog):
        """
        Univariate Local Linear Trend Model
        """
        # Model order
        k_states = k_posdef = 2

        # Initialize the statespace
        super(SemiLinearTrend, self).__init__(
            endog, k_states=k_states, k_posdef=k_posdef,
            initialization='approximate_diffuse',
            loglikelihood_burn=k_states)
        
        # Initialize the matrices
        self.ssm['design'] = np.array([1, 0])
        self.ssm['transition'] = np.array([[1, 1],
                                       [0, 1]])
        self.ssm['selection'] = np.eye(k_states)

        # Cache some indices
        self._state_cov_idx = ('state_cov',) + np.diag_indices(k_posdef)

    @property
    def param_names(self):
        return ['sigma2.measurement', 'sigma2.level', 'sigma2.trend', 'rho']

    @property
    def start_params(self):
        return np.r_[[np.nanstd(self.endog)]*3, 0.99]

    def transform_params(self, unconstrained):
        return unconstrained**2

    def untransform_params(self, constrained):
        return constrained**0.5

    def update(self, params, *args, **kwargs):
        params = super(SemiLinearTrend, self).update(params, *args, **kwargs)

        # Observation covariance
        self.ssm['obs_cov',0,0] = params[0]

        # State covariance
        self.ssm[self._state_cov_idx] = params[1:3]
        self['transition', 1, 1] = params[3]


class LocalLinearTrend(MLEModel):
    def __init__(self, endog, obs_cov=None):
        """
        Univariate Local Linear Trend Model
        """
        # Model order
        k_states = k_posdef = 2
        if obs_cov is not None:
            self.obs_cov_median = np.median(obs_cov)
            obs_cov = obs_cov[np.newaxis, np.newaxis, :]
           
            
        # Initialize the statespace
        super(LocalLinearTrend, self).__init__(
            endog, k_states=k_states, k_posdef=k_posdef,
            initialization='approximate_diffuse',
            loglikelihood_burn=k_states
        )
        # Initialize the matrices
        self.ssm['design'] = np.array([1, 0])
        self.ssm['transition'] = np.array([[1, 1],
                                       [0, 1]])
        self.ssm['selection'] = np.eye(k_states)

        # Cache some indices
        self._state_cov_idx = ('state_cov',) + np.diag_indices(k_posdef)
        if obs_cov is not None:
            self.ssm['obs_cov'] = obs_cov
        self.obs_cov = obs_cov
        
    def clone(self, endog, exog, **kwargs):
        # This method must be set to allow forecasting in custom
        # state space models that include time-varying state
        # space matrices, like we have for state_intercept here
        obs_cov = np.ones(endog.shape[0])*self.obs_cov_median
        mod = self.__class__(endog, obs_cov=obs_cov)
        return mod
    
    @property
    def param_names(self):
        if self.obs_cov is None:
            return ['sigma2.measurement', 'sigma2.level', 'sigma2.trend']
        return ['sigma2.level', 'sigma2.trend']

    @property
    def start_params(self):
        if self.obs_cov is None:
            return [np.nanstd(self.endog)]*3
        return [np.nanstd(self.endog)]*2

    def transform_params(self, unconstrained):
        return unconstrained**2

    def untransform_params(self, constrained):
        return constrained**0.5

    def update(self, params, *args, **kwargs):
        params = super(LocalLinearTrend, self).update(params, *args, **kwargs)
        
        if self.obs_cov is None:
            # Observation covariance
            self.ssm['obs_cov',0,0] = params[0]

            # State covariance
            self.ssm[self._state_cov_idx] = params[1:3]
        else:
            # State covariance
            self.ssm[self._state_cov_idx] = params[:]


class SO2FusionModel(MLEModel):
    def __init__(self, measurements, initial_state=0, initial_cov=1,
                 k_states=2, k_posdef=2, obs_cov=None):
        # Initialize the base MLEModel
        super().__init__(endog=measurements, k_states=k_states,
                         k_posdef=k_posdef)
        self.obs_cov = obs_cov
        # Initial state and covariance
        self.ssm.initialize_known(np.array(initial_state),
                                  np.eye(k_states)*initial_cov)

        # Define the state transition matrix (A)

        if k_states == 2:
            self.ssm['transition'] = np.array([[1, 1],
                                               [0, 1]])
        else:
            self.ssm.transition = np.array([[1]])

        # Define the selection matrix (B, for process noise)
        if k_states == 2:
            self.ssm['selection'] = np.eye(k_states)
        else:
            self.ssm.selection = np.array([[1]])

        # Define the measurement matrix (H)
        n_sensors = measurements.shape[1]
        if k_states == 2:
            self.ssm.design = np.tile([1, 0], (n_sensors, 1))
        else:
            self.ssm.design = np.tile([1], (n_sensors, 1))

        if self.obs_cov is not None:
            self.ssm.obs_cov = obs_cov

    def transform_params(self, unconstrained):
        return unconstrained**2

    def untransform_params(self, constrained):
        return constrained**0.5


    def update(self, params, *args, **kwargs):
        # Update the process noise covariance (Q)
        if self.ssm.k_states == 2:
            self.ssm.state_cov = np.diag(params[:2])
        else:
            self.ssm.state_cov = np.array([params[0]])
        # Update the measurement noise covariance (R)
        if self.obs_cov is None:
            self.ssm.obs_cov = np.diag(params[2:])



from datetime import datetime
import inspect
import os
import pandas as pd
from pathlib import Path
from functools import wraps
import hashlib
import logging

import numpy as np

logger = logging.getLogger(__name__)


def generate_cache_filename(func_name, args, kwargs):
    """
    Generate a unique filename for caching based on function name and arguments.
    """
    # Serialize arguments to a string representation
    instance = args[0] if args and hasattr(
        args[0], '__class__') else None

    if instance:
        args_repr = "_".join(repr(arg) for arg in args[1:])
    else:
        args_repr = "_".join(repr(arg) for arg in args)
    kwargs_repr = "_".join(f"{k}={v}" for k, v in sorted(kwargs.items()))

    # Combine everything into a single string
    combined = f"{func_name}_{args_repr}_{kwargs_repr}"

    # Hash the combined string to ensure the filename is valid and not too long
    hash_object = hashlib.md5(combined.encode("utf-8"))
    hashed_name = hash_object.hexdigest()

    # Construct the filename
    return f"{func_name}_{hashed_name}"  # Use .pkl for cached dataframes


def _is_class_method(func):
    """
    Check if a function is a class method.
    """
    func_params = list(inspect.signature(func).parameters)
    return func_params and func_params[0] == 'self'


def cache_dataframe(cache_dir: str = ""):
    """
    Decorator to cache pandas DataFrames, handle date ranges, and persist the cache to disk.

    Args:
        cache_key (str): The attribute name for the cached DataFrame.
        use_class_attrs (bool): If True, fetch start_date and end_date from the class instance if available.
        cache_file (str): Filepath for persisting the cache.
    """
    if not cache_dir:
        cache_dir = os.path.join(Path.home(), ".aitana_cache")
        os.makedirs(cache_dir, exist_ok=True)

    def decorator(func):
        @wraps(func)
        def wrapper(*args, start_date=None, end_date=None, clear_cache=False, **kwargs):
            cache_key = generate_cache_filename(func.__name__, args, kwargs)
            cache_file = os.path.join(cache_dir, f"{cache_key}.csv")
            if clear_cache:
                if os.path.exists(cache_file):
                    os.remove(cache_file)

            instance = args[0] if _is_class_method(func) else None

            if instance is not None:
                # Use instance attributes if not explicitly provided
                end_date = end_date or getattr(instance, 'end_date', None)
                start_date = start_date or getattr(
                    instance, 'start_date', None)

            if end_date is None:
                raise ValueError("end_date must be provided.")

            end_date = pd.to_datetime(end_date, utc=True)
            start_date = pd.to_datetime(start_date, utc=True)

            # Load the cache from disk if it exists
            if not hasattr(wrapper, 'cached_df'):
                if os.path.exists(cache_file):
                    logger.debug(f"Loading cache from {cache_file}")
                    cached_df = pd.read_csv(
                        cache_file, parse_dates=True, index_col=0,
                        date_format="ISO8601")
                else:
                    cached_df = pd.DataFrame()
                setattr(wrapper, 'cached_df', cached_df)
            else:
                cached_df = getattr(wrapper, 'cached_df')

            # Check the date range of the cached data
            if not cached_df.empty:
                cached_end = cached_df.index.max()
            else:
                cached_end = None

            # Determine missing ranges
            missing_end = end_date if cached_end is None or end_date > cached_end else None

            # If there are missing ranges, fetch the data
            if missing_end:
                logger.debug("Fetching missing data...")
                if instance is not None:
                    setattr(instance, 'end_date', missing_end)
                    missing_data = func(*args, **kwargs)
                else:
                    try:
                        missing_data = func(*args, end_date=end_date, **kwargs)
                    except TypeError:
                        missing_data = func(*args, **kwargs)

                if not isinstance(missing_data, pd.DataFrame):
                    raise ValueError(
                        "The decorated function must return a pandas DataFrame.")

                if not missing_data.empty:
                    missing_data.index = pd.to_datetime(
                        missing_data.index)  # Ensure the index is datetime

                # Update the cache
                cached_df = missing_data
                setattr(wrapper, 'cached_df', cached_df)

                # Persist the updated cache to disk
                logger.debug(f"Saving cache to {cache_file}")
                cached_df.to_csv(cache_file)
            # Return the relevant slice of the cached DataFrame
            return_df = cached_df.loc[start_date:end_date.replace(
                hour=23, minute=59, second=59)]
            if return_df.empty:
                raise ValueError("No data available for the given date range.")
            return return_df

        return wrapper
    return decorator


def gradient(df, period="14D"):
    """
    Compute gradient for time series by first smoothing
    the time series and then computing the first-order difference.

    Parameters:
    -----------
        :param df: Dataframe
        :type df: :class:`~pandas.DataFrame`
        :param period: Period over which to compute a rolling mean.
        :type period: str
    """
    df_smooth = df.rolling(period).mean()
    df_grad = df_smooth.diff()
    df_grad.loc[df_grad.index[0], "obs"] = 0.0
    return df_grad


def eqRate(cat, fixed_time=None, fixed_nevents=None, enddate=datetime.utcnow()):
    """
    Compute earthquake rate.

    :param cat: A catalogue of earthquakes as returned by
                :method:`pyvolprob.load_ruapehu_earthquakes`
    :type cat: :class:`pandas.DataFrame`
    :param fixed_time: If not 'None', compute the earthquake rate
                       based on a fixed-length time window given in
                       days.
    :type fixed_time: int
    :param fixed_nevents: If not None, compute the earthquake rate
                          based on a fixed number of events.
    :type fixed_nevents: int
    :param enddate: The latest date of the time-series.
                    Mainly needed for testing.
    :type enddate: :class:`datetime.datetime`

    """
    if fixed_time is not None and fixed_nevents is not None:
        raise ValueError(
            "Please define either 'fixed_time' or 'fixed_nevents'")

    dates = cat.index.values
    if fixed_time is not None:
        ds = pd.Series(np.ones(len(dates)), index=dates)
        ds = pd.concat([ds, pd.Series([np.nan], index=[enddate])])
        ds.sort_index(inplace=True)
        ds = ds.rolling("{:d}D".format(fixed_time)).count() / fixed_time
        ds.index -= pd.Timedelta("{:d}D".format(int(fixed_time / 2.0)))
        return pd.DataFrame({"obs": ds}).tz_localize('utc')
    elif fixed_nevents is not None:
        nevents = dates.shape[0]
        aBin = np.zeros(nevents - fixed_nevents, dtype="datetime64[ns]")
        aRate = np.zeros(nevents - fixed_nevents)
        iS = 0
        for s in np.arange(fixed_nevents, nevents):
            i1, i2 = s - fixed_nevents, s
            dt = (dates[i2] - dates[i1]).astype("timedelta64[s]")
            dt_days = dt.astype(float) / 86400.0
            aBin[iS] = dates[i1] + 0.5 * dt
            aRate[iS] = fixed_nevents / dt_days
            iS += 1
        return pd.DataFrame({"obs": aRate}, index=aBin).tz_localize('utc')
    else:
        raise ValueError(
            "Please define either 'fixed_time' or 'fixed_nevents'")


def reindex(df, dates, fill_method=None, ffill_interval=14):
    """
    Reindex and forward fill to generate a
    timeseries that can be used to set the evidence
    for a BN.

    :param df: Dataframe to reindex
    :type df: :class:`pandas.DataFrame`
    :param dates: new date index
    :type dates: :class:`pandas.DataTimeIndex`
    :param ffill_interval: Best-by interval for data
    :type ffill_interval: int
    """
    if fill_method is None:
        return df["obs"].resample("D").max().reindex(dates)
    elif fill_method == "ffill":
        return df["obs"].reindex(dates, method="ffill", limit=ffill_interval)
    elif fill_method == "interpolate":
        df_tmp = df["obs"].resample("D").max().reindex(dates)
        return df_tmp.interpolate(method="linear")
    else:
        msg = "'fill_method' has to be one of "
        msg += "[None, 'ffill', 'interpolate']"
        raise ValueError(msg)
